package api

import (
	"bytes"
	"context"
	"encoding/json"
	"io/ioutil"
	"net/http"
	"net/http/httptest"
	"strings"
	"testing"
	"time"

	"github.com/go-gorp/gorp"
	"github.com/golang/mock/gomock"
	"github.com/stretchr/testify/assert"
	"github.com/stretchr/testify/require"

	"github.com/ovh/cds/engine/api/application"
	"github.com/ovh/cds/engine/api/pipeline"
	"github.com/ovh/cds/engine/api/project"
	"github.com/ovh/cds/engine/api/repositoriesmanager"
	"github.com/ovh/cds/engine/api/services"
	"github.com/ovh/cds/engine/api/services/mock_services"
	"github.com/ovh/cds/engine/api/test"
	"github.com/ovh/cds/engine/api/test/assets"
	"github.com/ovh/cds/engine/api/workflow"
	"github.com/ovh/cds/sdk"
	"github.com/ovh/cds/sdk/log"
)

func TestPostUpdateWorkflowAsCodeHandler(t *testing.T) {
	api, db, _ := newTestAPI(t)

	_, _ = assets.InsertService(t, db, t.Name()+"_HOOKS", services.TypeHooks)
	_, _ = assets.InsertService(t, db, t.Name()+"_VCS", services.TypeVCS)
	_, _ = assets.InsertService(t, db, t.Name()+"_REPO", services.TypeRepositories)

	u, pass := assets.InsertAdminUser(t, db)

	UUID := sdk.UUID()

	svcs, errS := services.LoadAll(context.TODO(), db)
	assert.NoError(t, errS)
	for _, s := range svcs {
		_ = services.Delete(db, &s) // nolint
	}

	a, _ := assets.InsertService(t, db, "Test_postWorkflowAsCodeHandlerVCS", services.TypeVCS)
	b, _ := assets.InsertService(t, db, "Test_postWorkflowAsCodeHandlerRepo", services.TypeRepositories)
	c, _ := assets.InsertService(t, db, "Test_postWorkflowAsCodeHandlerHook", services.TypeHooks)

	defer func() {
		services.Delete(db, a)
		services.Delete(db, b)
		services.Delete(db, c)
	}()

	//This is a mock for the repositories service
	services.HTTPClient = mock(
		func(r *http.Request) (*http.Response, error) {
			body := new(bytes.Buffer)
			w := new(http.Response)
			enc := json.NewEncoder(body)
			w.Body = ioutil.NopCloser(body)
			w.StatusCode = http.StatusOK
			switch r.URL.String() {
			case "/operations":
				ope := new(sdk.Operation)
				ope.UUID = UUID
				ope.Status = sdk.OperationStatusProcessing
				if err := enc.Encode(ope); err != nil {
					return writeError(w, err)
				}
			case "/operations/" + UUID:
				ope := new(sdk.Operation)
				ope.UUID = UUID
				ope.Status = sdk.OperationStatusDone
				if err := enc.Encode(ope); err != nil {
					return writeError(w, err)
				}
			case "/vcs/github/repos/foo/myrepo":
				vcsRepo := sdk.VCSRepo{
					Name:         "foo/myrepo",
					SSHCloneURL:  "git:foo",
					HTTPCloneURL: "https:foo",
				}
				if err := enc.Encode(vcsRepo); err != nil {
					return writeError(w, err)
				}
			case "/vcs/github/repos/foo/myrepo/branches":
				bs := make([]sdk.VCSBranch, 1)
				bs[0] = sdk.VCSBranch{
					DisplayID: "master",
					Default:   true,
				}
				if err := enc.Encode(bs); err != nil {
					return writeError(w, err)
				}
			case "/vcs/github/webhooks":
				hookInfo := repositoriesmanager.WebhooksInfos{
					WebhooksSupported: true,
					WebhooksDisabled:  false,
				}
				if err := enc.Encode(hookInfo); err != nil {
					return writeError(w, err)
				}
			case "/vcs/github/repos/foo/myrepo/hooks":
				hook := sdk.VCSHook{
					ID: "myod",
				}
				if err := enc.Encode(hook); err != nil {
					return writeError(w, err)
				}
			case "/vcs/github/repos/foo/myrepo/pullrequests?state=open":
				vcsPRs := []sdk.VCSPullRequest{}
				if err := enc.Encode(vcsPRs); err != nil {
					return writeError(w, err)
				}
			case "/vcs/github/repos/foo/myrepo/pullrequests":
				vcsPR := sdk.VCSPullRequest{
					URL: "myURL",
				}
				if err := enc.Encode(vcsPR); err != nil {
					return writeError(w, err)
				}
			case "/task/bulk":
				var hooks map[string]sdk.NodeHook
				bts, err := ioutil.ReadAll(r.Body)
				if err != nil {
					return writeError(w, err)
				}
				if err := json.Unmarshal(bts, &hooks); err != nil {
					return writeError(w, err)
				}
				if err := enc.Encode(hooks); err != nil {
					return writeError(w, err)
				}
			default:
				t.Logf("[WRONG ROUTE] %s", r.URL.String())
				w.StatusCode = http.StatusNotFound
			}

			return w, nil
		},
	)

	assert.NoError(t, workflow.CreateBuiltinWorkflowHookModels(db))

	proj := createProject(t, db, api)
	pip := createPipeline(t, db, api, proj)
	app := createApplication(t, db, api, proj)

	repoModel, err := workflow.LoadHookModelByName(db, sdk.RepositoryWebHookModelName)
	assert.NoError(t, err)

	w := initWorkflow(t, db, proj, app, pip, repoModel)
	w.FromRepository = "myfromrepositoryurl"

	var errP error
	proj, errP = project.Load(api.mustDB(), proj.Key,
		project.LoadOptions.WithApplicationWithDeploymentStrategies,
		project.LoadOptions.WithPipelines,
		project.LoadOptions.WithEnvironments,
		project.LoadOptions.WithIntegrations,
	)
	assert.NoError(t, errP)
	if !assert.NoError(t, workflow.Insert(context.Background(), db, api.Cache, *proj, w)) {
		return
	}

	// Updating workflow
	w.WorkflowData.Node.Triggers = []sdk.NodeTrigger{
		{
			ChildNode: sdk.Node{
				Type: "fork",
				Name: "secondnode",
			},
		},
	}

	uri := api.Router.GetRoute("POST", api.postWorkflowAsCodeHandler, map[string]string{
		"key":              proj.Key,
		"permWorkflowName": w.Name,
	})

	req := assets.NewJWTAuthentifiedRequest(t, pass, "POST", uri, w)
	q := req.URL.Query()
	q.Set("branch", "master")
	q.Set("message", "my message")
	req.URL.RawQuery = q.Encode()

	// Do the request
	wr := httptest.NewRecorder()
	api.Router.Mux.ServeHTTP(wr, req)
	assert.Equal(t, 200, wr.Code)
	myOpe := new(sdk.Operation)
	test.NoError(t, json.Unmarshal(wr.Body.Bytes(), myOpe))
	assert.NotEmpty(t, myOpe.UUID)

	retry := 0
	for {
		retry++
		if retry > 10 {
			t.Log("Number of retry reached")
			t.Fail()
			break
		}

		// Get operation
		uriGET := api.Router.GetRoute("GET", api.getWorkflowAsCodeHandler, map[string]string{
			"key":              proj.Key,
			"permWorkflowName": w.Name,
			"uuid":             myOpe.UUID,
		})
		reqGET, err := http.NewRequest("GET", uriGET, nil)
		require.NoError(t, err)
		assets.AuthentifyRequest(t, reqGET, u, pass)
		wrGet := httptest.NewRecorder()
		api.Router.Mux.ServeHTTP(wrGet, reqGET)
		require.Equal(t, 200, wrGet.Code)

		myOpeGet := new(sdk.Operation)
		require.NoError(t, json.Unmarshal(wrGet.Body.Bytes(), myOpeGet))
		if myOpeGet.Status < sdk.OperationStatusDone {
			time.Sleep(1 * time.Second)
			continue
		}

		require.Equal(t, sdk.OperationStatusDone, myOpeGet.Status, "Invalid status for operation %v", string(wrGet.Body.Bytes()))
		assert.Equal(t, "myURL", myOpeGet.Setup.Push.PRLink)
		break
	}
}

func TestPostMigrateWorkflowAsCodeHandler(t *testing.T) {
	api, db, _ := newTestAPI(t)

	u, pass := assets.InsertAdminUser(t, db)

	UUID := sdk.UUID()

	a, _ := assets.InsertService(t, db, "Test_postWorkflowAsCodeHandlerVCS", services.TypeVCS)
	b, _ := assets.InsertService(t, db, "Test_postWorkflowAsCodeHandlerRepo", services.TypeRepositories)
	c, _ := assets.InsertService(t, db, "Test_postWorkflowAsCodeHandlerHook", services.TypeHooks)

	defer func() {
		_ = services.Delete(db, a)
		_ = services.Delete(db, b)
		_ = services.Delete(db, c)
	}()

	assert.NoError(t, workflow.CreateBuiltinWorkflowHookModels(db))

	//This is a mock for the repositories service
	services.HTTPClient = mock(
		func(r *http.Request) (*http.Response, error) {
			body := new(bytes.Buffer)
			w := new(http.Response)
			enc := json.NewEncoder(body)
			w.Body = ioutil.NopCloser(body)
			w.StatusCode = http.StatusOK
			switch r.URL.String() {
			case "/operations":
				ope := new(sdk.Operation)
				ope.UUID = UUID
				ope.Status = sdk.OperationStatusProcessing
				if err := enc.Encode(ope); err != nil {
					return writeError(w, err)
				}
			case "/operations/" + UUID:
				ope := new(sdk.Operation)
				ope.UUID = UUID
				ope.Status = sdk.OperationStatusDone
				if err := enc.Encode(ope); err != nil {
					return writeError(w, err)
				}
			case "/vcs/github/repos/foo/myrepo":
				vcsRepo := sdk.VCSRepo{
					Name:         "foo/myrepo",
					SSHCloneURL:  "git:foo",
					HTTPCloneURL: "https:foo",
				}
				if err := enc.Encode(vcsRepo); err != nil {
					return writeError(w, err)
				}
			case "/vcs/github/repos/foo/myrepo/branches":
				bs := make([]sdk.VCSBranch, 1)
				bs[0] = sdk.VCSBranch{
					DisplayID: "master",
					Default:   true,
				}
				if err := enc.Encode(bs); err != nil {
					return writeError(w, err)
				}
			case "/vcs/github/webhooks":
				hookInfo := repositoriesmanager.WebhooksInfos{
					WebhooksSupported: true,
					WebhooksDisabled:  false,
				}
				if err := enc.Encode(hookInfo); err != nil {
					return writeError(w, err)
				}
			case "/vcs/github/repos/foo/myrepo/hooks":
				hook := sdk.VCSHook{
					ID: "myod",
				}
				if err := enc.Encode(hook); err != nil {
					return writeError(w, err)
				}
			case "/vcs/github/repos/foo/myrepo/pullrequests?state=open":
				vcsPRs := []sdk.VCSPullRequest{}
				if err := enc.Encode(vcsPRs); err != nil {
					return writeError(w, err)
				}
			case "/vcs/github/repos/foo/myrepo/pullrequests":
				vcsPR := sdk.VCSPullRequest{
					URL: "myURL",
				}
				if err := enc.Encode(vcsPR); err != nil {
					return writeError(w, err)
				}
			case "/task/bulk":
				var hooks map[string]sdk.NodeHook
				bts, err := ioutil.ReadAll(r.Body)
				if err != nil {
					return writeError(w, err)
				}
				if err := json.Unmarshal(bts, &hooks); err != nil {
					return writeError(w, err)
				}
				for k, h := range hooks {
					if h.HookModelName == sdk.RepositoryWebHookModelName {
						cfg := hooks[k].Config
						cfg["webHookURL"] = sdk.WorkflowNodeHookConfigValue{
							Value:        "http://lolcat.host",
							Configurable: false,
						}
					}
				}
				if err := enc.Encode(hooks); err != nil {
					return writeError(w, err)
				}
			default:
				t.Logf("[WRONG ROUTE] %s", r.URL.String())
				w.StatusCode = http.StatusNotFound
			}

			return w, nil
		},
	)

	// Create Project
	proj := createProject(t, db, api)

	// Create Pipeline
	pip := createPipeline(t, db, api, proj)

	// Create Application
	app := createApplication(t, db, api, proj)

	repoModel, err := workflow.LoadHookModelByName(db, sdk.RepositoryWebHookModelName)
	assert.NoError(t, err)

	// Create Workflow
	w := initWorkflow(t, db, proj, app, pip, repoModel)

	var errP error
	proj, errP = project.Load(api.mustDB(), proj.Key,
		project.LoadOptions.WithApplicationWithDeploymentStrategies,
		project.LoadOptions.WithPipelines,
		project.LoadOptions.WithEnvironments,
		project.LoadOptions.WithIntegrations,
	)
	assert.NoError(t, errP)
	if !assert.NoError(t, workflow.Insert(context.Background(), db, api.Cache, *proj, w)) {
		return
	}

	t.Logf("%+v", w)

	uri := api.Router.GetRoute("POST", api.postWorkflowAsCodeHandler, map[string]string{
		"key":              proj.Key,
		"permWorkflowName": w.Name,
	})

	req := assets.NewJWTAuthentifiedRequest(t, pass, "POST", uri, nil)
	q := req.URL.Query()
	q.Set("migrate", "true")
	q.Set("branch", "master")
	q.Set("message", "my message")
	req.URL.RawQuery = q.Encode()

	// Do the request
	wr := httptest.NewRecorder()
	api.Router.Mux.ServeHTTP(wr, req)
	assert.Equal(t, 200, wr.Code)
	myOpe := new(sdk.Operation)
	test.NoError(t, json.Unmarshal(wr.Body.Bytes(), myOpe))
	assert.NotEmpty(t, myOpe.UUID)

	cpt := 0
	for {
		cpt++
		if cpt > 10 {
			t.Log("Number of retry reached")
			t.Fail()
			break
		}

		// Get operation
		uriGET := api.Router.GetRoute("GET", api.getWorkflowAsCodeHandler, map[string]string{
			"key":              proj.Key,
			"permWorkflowName": w.Name,
			"uuid":             myOpe.UUID,
		})
		reqGET, err := http.NewRequest("GET", uriGET, nil)
		require.NoError(t, err)
		assets.AuthentifyRequest(t, reqGET, u, pass)
		wrGet := httptest.NewRecorder()
		api.Router.Mux.ServeHTTP(wrGet, reqGET)
		require.Equal(t, 200, wrGet.Code)

		myOpeGet := new(sdk.Operation)
		require.NoError(t, json.Unmarshal(wrGet.Body.Bytes(), myOpeGet))
		if myOpeGet.Status < sdk.OperationStatusDone {
			time.Sleep(1 * time.Second)
			continue
		}

		require.Equal(t, sdk.OperationStatusDone, myOpeGet.Status, "Invalid status for operation %v", string(wrGet.Body.Bytes()))
		assert.Equal(t, "myURL", myOpeGet.Setup.Push.PRLink)
		break
	}
}

func createProject(t *testing.T, db *gorp.DbMap, api *API) *sdk.Project {
	// Create Project
	pkey := sdk.RandomString(10)
	proj := assets.InsertTestProject(t, db, api.Cache, pkey, pkey)
	vcsServer := sdk.ProjectVCSServerLink{
		ProjectID: proj.ID,
		Name:      "github",
	}
	vcsServer.Set("token", "foo")
	vcsServer.Set("secret", "bar")
	assert.NoError(t, repositoriesmanager.InsertProjectVCSServerLink(context.TODO(), db, &vcsServer))

	return proj
}

func createPipeline(t *testing.T, db gorp.SqlExecutor, api *API, proj *sdk.Project) *sdk.Pipeline {
	pip := sdk.Pipeline{
		Name:      sdk.RandomString(10),
		ProjectID: proj.ID,
	}
	assert.NoError(t, pipeline.InsertPipeline(db, &pip))
	return &pip
}

func createApplication(t *testing.T, db gorp.SqlExecutor, api *API, proj *sdk.Project) *sdk.Application {
	app := sdk.Application{
		Name:               sdk.RandomString(10),
		ProjectID:          proj.ID,
		RepositoryFullname: "foo/myrepo",
		VCSServer:          "github",
	}
	assert.NoError(t, application.Insert(db, *proj, &app))
	assert.NoError(t, repositoriesmanager.InsertForApplication(db, &app, proj.Key))
	return &app
}

func initWorkflow(t *testing.T, db gorp.SqlExecutor, proj *sdk.Project, app *sdk.Application, pip *sdk.Pipeline, repoModel *sdk.WorkflowHookModel) *sdk.Workflow {
	w := sdk.Workflow{
		ProjectID:  proj.ID,
		ProjectKey: proj.Key,
		Name:       sdk.RandomString(10),
		WorkflowData: sdk.WorkflowData{
			Node: sdk.Node{
				Name: "root",
				Type: sdk.NodeTypePipeline,
				Context: &sdk.NodeContext{
					PipelineID:    pip.ID,
					ApplicationID: app.ID,
				},
				Hooks: []sdk.NodeHook{
					{
						HookModelName: sdk.RepositoryWebHookModelName,
						UUID:          sdk.RandomString(10),
						Config:        sdk.RepositoryWebHookModel.DefaultConfig.Clone(),
						HookModelID:   repoModel.ID,
					},
				},
			},
		},
	}
	w.WorkflowData.Node.Hooks[0].Config["webHookURL"] = sdk.WorkflowNodeHookConfigValue{Type: "string", Value: "http://ovh/com"}
	assert.NoError(t, workflow.RenameNode(context.Background(), db, &w))
	return &w
}

func Test_WorkflowAsCodeWithNotifications(t *testing.T) {
	api, db, _ := newTestAPI(t)

	_, _ = assets.InsertService(t, db, t.Name()+"_HOOKS", services.TypeHooks)
	_, _ = assets.InsertService(t, db, t.Name()+"_VCS", services.TypeVCS)
	_, _ = assets.InsertService(t, db, t.Name()+"_REPO", services.TypeRepositories)

	// Setup a mock for all services called by the API
	ctrl := gomock.NewController(t)
	defer ctrl.Finish()
	// The mock has been geenrated by mockgen: go get github.com/golang/mock/mockgen
	// If you have to regenerate thi mock you just have to run, from directory $GOPATH/src/github.com/ovh/cds/engine/api/services:
	// mockgen -source=http.go -destination=mock_services/services_mock.go Client
	servicesClients := mock_services.NewMockClient(ctrl)
	services.NewClient = func(_ gorp.SqlExecutor, _ []sdk.Service) services.Client {
		return servicesClients
	}
	defer func() {
		services.NewClient = services.NewDefaultClient
	}()

	// Create a project with a repository manager
	prjKey := sdk.RandomString(10)
	proj := assets.InsertTestProject(t, db, api.Cache, prjKey, prjKey)
	u, pass := assets.InsertLambdaUser(t, db, &proj.ProjectGroups[0].Group)

	vcsServer := sdk.ProjectVCSServerLink{
		ProjectID: proj.ID,
		Name:      "github",
	}
	vcsServer.Set("token", "foo")
	vcsServer.Set("secret", "bar")
	assert.NoError(t, repositoriesmanager.InsertProjectVCSServerLink(context.TODO(), db, &vcsServer))

	// Perform a "import as-code operation" to create a new workflow
	ope := `{"repo_fullname":"fsamin/go-repo",  "vcs_server": "github", "url":"https://github.com/fsamin/go-repo.git","strategy":{"connection_type":"https","ssh_key":"","user":"","password":"","branch":"","default_branch":"master","pgp_key":""},"setup":{"checkout":{"branch":"master"}}}`
	uri := api.Router.GetRoute("POST", api.postImportAsCodeHandler, map[string]string{
		"permProjectKey": proj.Key,
	})

	UUID := sdk.UUID()

	servicesClients.EXPECT().
		DoJSONRequest(gomock.Any(), "POST", "/operations", gomock.Any(), gomock.Any()).
		Return(nil, 201, nil).Times(2)

	servicesClients.EXPECT().
		DoJSONRequest(gomock.Any(), "GET", "/vcs/github/repos/fsamin/go-repo", gomock.Any(), gomock.Any(), gomock.Any()).MinTimes(0)

	servicesClients.EXPECT().
		DoJSONRequest(gomock.Any(), "GET", "/vcs/github/repos/fsamin/go-repo/branches", gomock.Any(), gomock.Any(), gomock.Any()).
		DoAndReturn(
			func(ctx context.Context, method, path string, in interface{}, out interface{}, _ interface{}) (http.Header, int, error) {
				bs := []sdk.VCSBranch{}
				b := sdk.VCSBranch{
					DisplayID:    "master",
					LatestCommit: "aaaaaaa",
				}
				bs = append(bs, b)
				out = bs
				return nil, 200, nil
			},
		).MaxTimes(10)

	servicesClients.EXPECT().
		DoJSONRequest(gomock.Any(), "GET", "/vcs/github/repos/fsamin/go-repo/branches/?branch=master", gomock.Any(), gomock.Any(), gomock.Any()).
		DoAndReturn(
			func(ctx context.Context, method, path string, in interface{}, out interface{}, _ interface{}) (http.Header, int, error) {
				b := &sdk.VCSBranch{
					DisplayID:    "master",
					LatestCommit: "aaaaaaa",
				}
				*(out.(*sdk.VCSBranch)) = *b
				return nil, 200, nil
			},
		).MaxTimes(10)

	servicesClients.EXPECT().
		DoJSONRequest(gomock.Any(), "GET", "/vcs/github/repos/fsamin/go-repo/commits/aaaaaaa", gomock.Any(), gomock.Any(), gomock.Any()).
		DoAndReturn(
			func(ctx context.Context, method, path string, in interface{}, out interface{}, _ interface{}) (http.Header, int, error) {
				c := &sdk.VCSCommit{
					Author: sdk.VCSAuthor{
						Avatar:      "me",
						DisplayName: "me",
						Name:        "me",
					},
					Hash:    "aaaaaaa",
					Message: "this is it",
				}
				*(out.(*sdk.VCSCommit)) = *c
				return nil, 200, nil
			},
		).MaxTimes(10)

	servicesClients.EXPECT().
		DoJSONRequest(gomock.Any(), "GET", "/vcs/github/repos/fsamin/go-repo/commits/aaaaaaa/statuses", gomock.Any(), gomock.Any(), gomock.Any()).
		DoAndReturn(
			func(ctx context.Context, method, path string, in interface{}, out interface{}, _ interface{}) (http.Header, int, error) {
				return nil, 200, nil
			},
		).MaxTimes(10)

	servicesClients.EXPECT().
		DoJSONRequest(gomock.Any(), "GET", "/vcs/github", gomock.Any(), gomock.Any()).
		DoAndReturn(
			func(ctx context.Context, method, path string, in interface{}, out interface{}) (http.Header, int, error) {
				return nil, 200, nil
			},
		).MaxTimes(10)

	servicesClients.EXPECT().
		DoJSONRequest(gomock.Any(), "POST", "/vcs/github/status", gomock.Any(), gomock.Any(), gomock.Any()).
		DoAndReturn(
			func(ctx context.Context, method, path string, in interface{}, out interface{}, _ interface{}) (http.Header, int, error) {
				return nil, 200, nil
			},
		).MaxTimes(10)

	servicesClients.EXPECT().
		DoJSONRequest(gomock.Any(), "GET", gomock.Any(), gomock.Any(), gomock.Any()).
		DoAndReturn(
			func(ctx context.Context, method, path string, in interface{}, out interface{}) (http.Header, int, error) {
				ope := new(sdk.Operation)
				ope.URL = "https://github.com/fsamin/go-repo.git"
				ope.UUID = UUID
				ope.Status = sdk.OperationStatusDone
				ope.VCSServer = "github"
				ope.RepoFullName = "fsamin/go-repo"
				ope.RepositoryStrategy.Branch = "master"
				ope.Setup.Checkout.Branch = "master"
				ope.RepositoryInfo = new(sdk.OperationRepositoryInfo)
				ope.RepositoryInfo.Name = "fsamin/go-repo"
				ope.RepositoryInfo.DefaultBranch = "master"
				ope.RepositoryInfo.FetchURL = "https://github.com/fsamin/go-repo.git"
				ope.LoadFiles.Pattern = workflow.WorkflowAsCodePattern
				ope.LoadFiles.Results = map[string][]byte{
					"w-go-repo.yml": []byte(`
name: w-go-repo
version: v2.0
workflow:
  playbook-game-molecule:
    depends_on:
    - root
    when:
    - success
    pipeline: build
    application: go-repo
  playbook-game-stuff:
    depends_on:
    - root
    when:
    - success
    pipeline: build
    application: go-repo
  netauto-versioning:
    depends_on:
    - release-prerequisites
    - playbook-game-molecule
    - playbook-game-stuff
    when:
    - success
    pipeline: build
    application: go-repo
  notify:
    depends_on:
    - netauto-versioning
    pipeline: build
    application: go-repo
  release-prerequisites:
    depends_on:
    - root
    when:
    - success
    pipeline: build
    application: go-repo
  root:
    pipeline: build
    application: go-repo
metadata:
  default_tags: tag_version,git.branch,git.author
notifications:
- type: vcs
- type: jabber
  pipelines:
  - playbook-game-stuff
  - playbook-game-molecule
  - netauto-versioning
  - release-prerequisites
  settings:
    on_success: never
    recipients:
    - ""
    template:
      subject: '{{.cds.project}}/{{.cds.workflow}}/{{.cds.pipeline}}#{{.cds.version}}
        {{.cds.status}}'
      body: |
        {{.cds.buildURL}}
- type: jabber
  pipelines:
  - notify
  settings:
    on_success: always
    on_failure: never
    recipients:
    - conf.netauto-release-mgmt@conference-2-standaloneclustere6001.corp.ovh.com
    template:
      subject: New release for {{.cds.workflow}}
      body: '{{.cds.workflow}} is now at version {{.workflow.netauto-versioning.build.VERSION}}'
    conditions:
      check:
      - variable: git.branch
        operator: eq
        value: master
- type: jabber
  pipelines:
  - notify
  settings:
    on_success: always
    on_failure: never

`),
					"go-repo.app.yml": []byte(`
name: go-repo
version: v1.0
repo: fsamin/go-repo
vcs_server: github
`),
					"go-repo.pip.yml": []byte(`name: build
version: v1.0`),
				}
				*(out.(*sdk.Operation)) = *ope
				return nil, 200, nil
			},
		).Times(2)

	servicesClients.EXPECT().
		DoJSONRequest(gomock.Any(), "POST", "/task/bulk", gomock.Any(), gomock.Any()).
		DoAndReturn(
			func(ctx context.Context, method, path string, in interface{}, out interface{}) (http.Header, int, error) {
				actualHooks, ok := in.(map[string]sdk.NodeHook)
				require.True(t, ok)
				require.Len(t, actualHooks, 1)
				for k, h := range actualHooks {
					assert.Equal(t, "RepositoryWebHook", h.HookModelName)
					assert.Equal(t, "POST", h.Config["method"].Value)
					assert.Equal(t, "fsamin/go-repo", h.Config["repoFullName"].Value)
					assert.Equal(t, "github", h.Config["vcsServer"].Value)
					assert.Equal(t, "w-go-repo", h.Config["workflow"].Value)
					h.Config["webHookURL"] = sdk.WorkflowNodeHookConfigValue{
						Value:        "http://lolcat.host",
						Configurable: false,
					}
					actualHooks[k] = h
				}
				out = actualHooks
				return nil, 200, nil
			},
		)

	servicesClients.EXPECT().
		DoJSONRequest(gomock.Any(), "GET", "/vcs/github/webhooks", gomock.Any(), gomock.Any(), gomock.Any()).
		DoAndReturn(
			func(ctx context.Context, method, path string, in interface{}, out interface{}, _ interface{}) (http.Header, int, error) {
				*(out.(*repositoriesmanager.WebhooksInfos)) = repositoriesmanager.WebhooksInfos{
					WebhooksSupported: true,
					WebhooksDisabled:  false,
					Icon:              sdk.GitHubIcon,
					Events: []string{
						"push",
					},
				}

				return nil, 200, nil
			},
		)

	servicesClients.EXPECT().
		DoJSONRequest(gomock.Any(), "POST", "/vcs/github/repos/fsamin/go-repo/grant", gomock.Any(), gomock.Any(), gomock.Any())

	servicesClients.EXPECT().
		DoJSONRequest(gomock.Any(), "POST", "/vcs/github/repos/fsamin/go-repo/hooks", gomock.Any(), gomock.Any(), gomock.Any()).
		DoAndReturn(
			func(ctx context.Context, method, path string, in interface{}, out interface{}, _ interface{}) (http.Header, int, error) {
				log.Debug("--> %T %+v", in, in)

				vcsHooks, ok := in.(*sdk.VCSHook)
				require.True(t, ok)

				require.Len(t, vcsHooks.Events, 0, "events list should be empty, default value is set by vcs")
				vcsHooks.Events = []string{"push"}

				assert.Equal(t, "POST", vcsHooks.Method)
				assert.Equal(t, "http://lolcat.host", vcsHooks.URL)
				vcsHooks.ID = sdk.UUID()
				*(out.(*sdk.VCSHook)) = *vcsHooks

				return nil, 200, nil
			},
		)

	req, err := http.NewRequest("POST", uri, strings.NewReader(ope))
	require.NoError(t, err)
	assets.AuthentifyRequest(t, req, u, pass)

	// Do the request
	w := httptest.NewRecorder()
	api.Router.Mux.ServeHTTP(w, req)
	assert.Equal(t, 201, w.Code)

	uri = api.Router.GetRoute("POST", api.postPerformImportAsCodeHandler, map[string]string{
		"permProjectKey": prjKey,
		"uuid":           UUID,
	})
	req, err = http.NewRequest("POST", uri, nil)
	require.NoError(t, err)
	assets.AuthentifyRequest(t, req, u, pass)

	// Do the request
	w = httptest.NewRecorder()
	api.Router.Mux.ServeHTTP(w, req)
	assert.Equal(t, 200, w.Code)
	t.Logf(w.Body.String())

	wk, err := workflow.Load(context.Background(), db, api.Cache, *proj, "w-go-repo", workflow.LoadOptions{})
	assert.NoError(t, err)
	assert.NotNil(t, wk)

	require.Len(t, wk.Notifications, 4, "not the right number of notifications")

	// Then we will trigger a run of the workflow wich should trigger an as-code operation
	vars := map[string]string{
		"key":              proj.Key,
		"permWorkflowName": wk.Name,
	}
	uri = api.Router.GetRoute("POST", api.postWorkflowRunHandler, vars)
	require.NotEmpty(t, uri)

	opts := &sdk.WorkflowRunPostHandlerOption{
		Hook: &sdk.WorkflowNodeRunHookEvent{
			WorkflowNodeHookUUID: wk.WorkflowData.Node.Hooks[0].UUID,
			Payload: map[string]string{
				"git.branch": "master",
			},
		},
	}
	req = assets.NewAuthentifiedRequest(t, u, pass, "POST", uri, opts)

	//Do the request
	rec := httptest.NewRecorder()
	api.Router.Mux.ServeHTTP(rec, req)
	assert.Equal(t, 202, rec.Code)

	if rec.Code != 202 {
		t.Logf("body => %s", rec.Body.String())
		t.FailNow()
	}

	var wrun sdk.WorkflowRun
	require.NoError(t, json.Unmarshal(rec.Body.Bytes(), &wrun))

	require.NoError(t, waitCraftinWorkflow(t, api.mustDB(), wrun.ID))
	wr, _ := workflow.LoadRunByID(db, wrun.ID, workflow.LoadRunOptions{})

	assert.NotEqual(t, "Fail", wr.Status)

	wk, err = workflow.Load(context.Background(), db, api.Cache, *proj, "w-go-repo", workflow.LoadOptions{})
	assert.NoError(t, err)
	assert.NotNil(t, wk)

	require.Len(t, wk.Notifications, 4, "not the right number of notifications")
}
